package org.rayflood.mikuvat.io;

import java.beans.XMLDecoder;
import java.beans.XMLEncoder;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.io.Serializable;
import java.util.Arrays;
import java.util.Comparator;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;
import java.util.zip.ZipOutputStream;

public abstract class ImmutableObject implements Serializable, Cloneable, Comparable<ImmutableObject>{
	private static final long serialVersionUID = 768886758398033337L;

	public abstract ImmutableObject setDefault();

	public abstract ImmutableObject delete();

	public abstract boolean isDeleted();

	public ImmutableObject cover(Object obj){
		return over(obj, true, false);
	}

	public ImmutableObject heap(Object obj){
		return over(obj, false, true);
	}

	public ImmutableObject lay(Object obj){
		return over(obj, false, false);
	}

	public abstract ImmutableObject over(Object obj, boolean fill, boolean has);

	public abstract String toString();

	public abstract boolean isReadied();

	public boolean equals(Object obj){
		if(obj == null || !(obj instanceof ImmutableObject) ||  !getClass().getName().equals(obj.getClass().getName())){
			return false;
		}
		try{
			return Arrays.equals(encodeBytes(), ((ImmutableObject)obj).encodeBytes());
		}
		catch(Exception e){
			throw new RuntimeException(e);
		}
	}

	public abstract int compareTo(ImmutableObject obj);

	protected static int compare(Comparable o1, Comparable o2){
		if(o1 == null && o2 == null){
			return 0;
		}
		if(o1 == null){
			return -1;
		}
		if(o2 == null){
			return 1;
		}
		return o1.compareTo(o2);
	}

	public static Comparator<ImmutableObject> getComparator(){
		return new Comparator<ImmutableObject>(){
			public int compare(ImmutableObject o1, ImmutableObject o2){
				return o1.compareTo(o2);
			}
		};
	}

	public static Comparator<ImmutableObject> getComparatorR(){
		return new Comparator<ImmutableObject>(){
			public int compare(ImmutableObject o1, ImmutableObject o2){
				return o1.compareTo(o2) * -1;
			}
		};
	}

	protected Object clone(){
		try{
			return decodeBytes(encodeBytes());
		}
		catch(Exception e){
			throw new RuntimeException(e);
		}
	}

	private static Object decodeBytes(byte[] buf) throws Exception{
		return unserialize(new ByteArrayInputStream(buf));
	}

	private static Object unserialize(InputStream in) throws Exception{
		ObjectInputStream ois = new ObjectInputStream(in);
		return ois.readObject();
	}

	private byte[] encodeBytes() throws Exception{
		ByteArrayOutputStream byteOut = new ByteArrayOutputStream();
		serialize(byteOut);
		return byteOut.toByteArray();
	}

	private void serialize(OutputStream out) throws Exception{
		ObjectOutputStream oos = new ObjectOutputStream(out);
		oos.writeObject(this);
		oos.close();
	}

	protected static Object load(String file) throws Exception{
		return load(new File(file));
	}

	protected static Object load(File file) throws Exception{
		return unserialize(new BufferedInputStream(new FileInputStream(file)));
	}

	protected void store(String file) throws Exception{
		store(new File(file));
	}

	protected void store(File file) throws Exception{
		serialize(new BufferedOutputStream(new FileOutputStream(file)));
	}

	protected static Object unzip(String file) throws Exception{
		return unzip(new File(file));
	}

	protected static Object unzip(File file) throws Exception{
		ZipInputStream zis = new ZipInputStream(new BufferedInputStream(new FileInputStream(file)));
		zis.getNextEntry();
		return unserialize(zis);
	}

	protected void zip(String file) throws Exception{
		zip(new File(file));
	}

	protected void zip(File file) throws Exception{
		ZipOutputStream zos = new ZipOutputStream(new BufferedOutputStream(new FileOutputStream(file)));
		zos.putNextEntry(new ZipEntry(getClass().getName()));
		serialize(zos);
	}

	protected static Object decodeXML(String xml) throws Exception{
		return decodeXML(new File(xml));
	}

	protected static Object decodeXML(File xml) throws Exception{
		XMLDecoder d = new XMLDecoder(new BufferedInputStream(new FileInputStream(xml)));
		Object result = d.readObject();
		d.close();
		return result;
	}

	protected void encodeXML(String xml) throws Exception{
		encodeXML(new File(xml));
	}

	protected void encodeXML(File xml) throws Exception{
		XMLEncoder e = new XMLEncoder(new BufferedOutputStream(new FileOutputStream(xml)));
		e.writeObject(this);
		e.close();
	}
}
