package org.rayflood.mikuvat.io;

import java.io.BufferedReader;
import java.io.StringReader;
import java.math.BigDecimal;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.regex.Pattern;

import javax.sound.midi.MetaMessage;
import javax.sound.midi.MidiEvent;
import javax.sound.midi.MidiMessage;
import javax.sound.midi.Track;

public class VSQDecoder{
	public static final Pattern SPLITEQUAL = Pattern.compile("=");
	public static final Pattern SPLITCOMMA = Pattern.compile(",");
	public static final Pattern SPLITDQUOTE = Pattern.compile("\\\"");
	public static final Pattern MMHEADER = Pattern.compile("^DM:\\d+:$");
	public static final Pattern SECTIONLINE = Pattern.compile("^\\[.+\\]$");
	public static final Charset WIN31J = Charset.forName("Windows-31J");
	public static final int TEXTLENGTH = 119;

	public static String[] splitEqual(String s){
		return SPLITEQUAL.split(s, 2);
	}

	public static String setSectionBracket(String s){
		return "[" + s + "]";
	}

	public static Integer adjustMaxMin(Integer value, Integer max, Integer min){
		if(value != null){
			if(value > max){
				value = max;
			}
			else if(value < min){
				value = min;
			}
		}
		return value;
	}

	public static BigDecimal adjustMaxMin(BigDecimal value, BigDecimal max, BigDecimal min){
		if(value != null){
			if(value.floatValue() > max.floatValue()){
				value = max;
			}
			else if(value.floatValue() < min.floatValue()){
				value = min;
			}
		}
		return value;
	}

	public static String getStringValue(Boolean value){
		if(value == null || !value){
			return "0";
		}
		else{
			return "1";
		}
	}

	public static boolean getBooleanValue(String value){
		if(value == null || value.isEmpty() || "0".equals(value)){
			return false;
		}
		else{
			return true;
		}
	}

	public static String getBigDecimalString(BigDecimal bd){
		return getBigDecimalString(bd, "%.6f");
	}

	public static String getBigDecimalString(BigDecimal bd, String format){
		return String.format(format, bd);
	}

	public static String getByteArrayString(byte[] data){
		StringBuffer sb = new StringBuffer();
		if(data != null && data.length > 0){
			sb.append(String.format("%X", data[0]));
			for(int i = 1; i < data.length; i++){
				sb.append(" " + String.format("%X", data[i]));
			}
		}
		return sb.toString();
	}

	public static Map<String, List<String>> decode(Track tr) throws Exception{
		List<Byte> v = new ArrayList<Byte>();
		for(int i = 1; i < tr.size(); i++){
			MidiMessage m = tr.get(i).getMessage();
			if(m instanceof MetaMessage){
				byte b[] = ((MetaMessage)m).getData();
				if(b.length == 0){
					continue;
				}
				int h = 1;
				String mmheader;
				do{
					h++;
					mmheader = new String(b, 0, h * 4, WIN31J);
				}while(!MMHEADER.matcher(mmheader).matches());
				for(int k = h * 4; k < b.length; k++){
					if(b[k] != 0x00){
						v.add(new Byte(b[k]));
					}
				}
			}
		}
		byte b[] = new byte[v.size()];
		for(int i = 0; i < v.size(); i++){
			b[i] = v.get(i);
		}
		String textevent = new String(b, WIN31J);

		Map<String, List<String>> sections = new LinkedHashMap<String, List<String>>();
		BufferedReader br = new BufferedReader(new StringReader(textevent));
		String bf;
		String sn = "";
		List<String> sv = new ArrayList<String>();
		do{
			bf = br.readLine();
			if(bf == null || SECTIONLINE.matcher(bf).matches()){
				if(sn.length() > 0 && !sections.containsKey(sn)){
					sections.put(sn.substring(1, sn.length() - 1), sv);
				}
				sn = bf;
				sv = new ArrayList<String>();
			}
			else if(bf.length() > 0){
				sv.add(bf);
			}
		}while(bf != null);

		return sections;
	}

	public static List<MidiEvent> encode(List<String> trackSections) throws Exception{
		String te = "";
		for(int i = 0; i < trackSections.size(); i++){
			String vts = trackSections.get(i);
			te += vts + "\n";
		}

		List<MidiEvent> tme = new ArrayList<MidiEvent>();
		byte b[] = te.getBytes(WIN31J);
		for(int j = 0; j <= b.length / TEXTLENGTH; j++){
			int h = (int)StrictMath.ceil((double)Integer.toString(j).length() / 4.0);
			String mef = "DM:%0" + h * 4 + "d:";
			byte header[] = String.format(mef, j).getBytes(WIN31J);
			List<Byte> mbl = new ArrayList<Byte>();
			for(int k = 0; k < header.length; k++){
				mbl.add(header[k]);
			}
			for(int k = 0; k < TEXTLENGTH; k++){
				int bi = j * TEXTLENGTH + k;
				if(bi >= b.length){
					break;
				}
				mbl.add(b[bi]);
			}
			byte mb[] = new byte[mbl.size()];
			for(int k = 0; k < mbl.size(); k++){
				mb[k] = mbl.get(k);
			}
			MetaMessage mm = new MetaMessage();
			mm.setMessage(1, mb, mb.length);
			tme.add(new MidiEvent(mm, 0L));
		}

		return tme;
	}

	public static List<VibratoBP> getVibratoBPList(SortedMap<BigDecimal, Integer> bpm){
		List<VibratoBP> bpl = new ArrayList<VibratoBP>();
		BigDecimal xi[] = bpm.keySet().toArray(new BigDecimal[0]);
		for(int i = 0; i < xi.length; i++){
			BigDecimal x = xi[i];
			int y = bpm.get(x);
			bpl.add(new VibratoBP(x, y));
		}
		return bpl;
	}

	public static TreeMap<BigDecimal, Integer> getVibratoBPMap(List<VibratoBP> bpl){
		TreeMap<BigDecimal, Integer> bpm = new TreeMap<BigDecimal, Integer>();
		for(int i = 0; i < bpl.size(); i++){
			VibratoBP bp = bpl.get(i);
			bpm.put(bp.getX(), bp.getY());
		}
		return bpm;
	}
}
